@@ -18,6 +18,8 @@ package v1
 
 import (
 	"errors"
+	"fmt"
+	"os"
 	"strings"
 
 	corev1 "k8s.io/api/core/v1"
@@ -30,30 +32,46 @@ import (
 
 type QuayVersion string
 
+var QuayVersionCurrent QuayVersion = QuayVersion(os.Getenv("QUAY_VERSION"))
+
+type ComponentKind string
+
 const (
-	QuayVersionCurrent  QuayVersion = "vader"
-	QuayVersionPrevious QuayVersion = ""
+	ComponentBase          ComponentKind = "base"
+	ComponentPostgres                    = "postgres"
+	ComponentClair                       = "clair"
+	ComponentRedis                       = "redis"
+	ComponentHPA                         = "horizontalpodautoscaler"
+	ComponentObjectStorage               = "objectstorage"
+	ComponentRoute                       = "route"
+	ComponentMirror                      = "mirror"
+	ComponentMonitoring                  = "monitoring"
+	ComponentTLS                         = "tls"
 )
 
-// CanUpgrade returns true if the given version can be upgraded to this Operator's synchronized Quay version.
-func CanUpgrade(fromVersion QuayVersion) bool {
-	return fromVersion == QuayVersionCurrent || fromVersion == QuayVersionPrevious
+var allComponents = []ComponentKind{
+	ComponentPostgres,
+	ComponentClair,
+	ComponentRedis,
+	ComponentHPA,
+	ComponentObjectStorage,
+	ComponentRoute,
+	ComponentMirror,
+	ComponentMonitoring,
+	ComponentTLS,
 }
 
-var allComponents = []string{
-	"postgres",
-	"clair",
-	"redis",
-	"horizontalpodautoscaler",
-	"objectstorage",
-	"route",
-	"mirror",
+var requiredComponents = []ComponentKind{
+	ComponentPostgres,
+	ComponentObjectStorage,
+	ComponentRoute,
+	ComponentRedis,
 }
 
-var requiredComponents = []string{
-	"postgres",
-	"objectstorage",
-}
+const (
+	ManagedKeysName         = "quay-registry-managed-secret-keys"
+	QuayConfigTLSSecretName = "quay-config-tls"
+)
 
 // QuayRegistrySpec defines the desired state of QuayRegistry.
 type QuayRegistrySpec struct {
@@ -66,7 +84,7 @@ type QuayRegistrySpec struct {
 // Component describes how the Operator should handle a backing Quay service.
 type Component struct {
 	// Kind is the unique name of this type of component.
-	Kind string `json:"kind"`
+	Kind ComponentKind `json:"kind"`
 	// Managed indicates whether or not the Operator is responsible for the lifecycle of this component.
 	// Default is true.
 	Managed bool `json:"managed"`
@@ -74,24 +92,40 @@ type Component struct {
 
 type ConditionType string
 
+// Follow a list of known condition types. These are used when reporting a QuayRegistry status
+// through its .status.conditions slice.
 const (
 	ConditionTypeAvailable      ConditionType = "Available"
 	ConditionTypeRolloutBlocked ConditionType = "RolloutBlocked"
-	// TODO: Add more useful conditions.
+	ConditionComponentsCreated  ConditionType = "ComponentsCreated"
+	ComponentBaseReady          ConditionType = "ComponentBaseReady"
+	ComponentPostgresReady      ConditionType = "ComponentPostgresReady"
+	ComponentClairReady         ConditionType = "ComponentClairReady"
+	ComponentRedisReady         ConditionType = "ComponentRedisReady"
+	ComponentHPAReady           ConditionType = "ComponentHPAReady"
+	ComponentObjectStorageReady ConditionType = "ComponentObjectStorageReady"
+	ComponentRouteReady         ConditionType = "ComponentRouteReady"
+	ComponentMirrorReady        ConditionType = "ComponentMirrorReady"
+	ComponentMonitoringReady    ConditionType = "ComponentMonitoringReady"
+	ComponentTLSReady           ConditionType = "ComponentTLSReady"
 )
 
 type ConditionReason string
 
+// Below follow a list of all Reasons used while reporting QuayRegistry conditions through its
+// .status.conditions field.
 const (
-	// ConditionTypeAvailable
-	ConditionReasonHealthChecksPassing  ConditionReason = "HealthChecksPassing"
-	ConditionReasonMigrationsInProgress ConditionReason = "MigrationsInProgress"
-	// ConditionTypeRolloutBlocked
+	ConditionReasonComponentNotReady                     ConditionReason = "ComponentNotReady"
+	ConditionReasonComponentReady                        ConditionReason = "ComponentReady"
+	ConditionReasonComponentUnmanaged                    ConditionReason = "ComponentNotManaged"
+	ConditionReasonHealthChecksPassing                   ConditionReason = "HealthChecksPassing"
+	ConditionReasonMigrationsInProgress                  ConditionReason = "MigrationsInProgress"
 	ConditionReasonComponentsCreationSuccess             ConditionReason = "ComponentsCreationSuccess"
 	ConditionReasonUpgradeUnsupported                    ConditionReason = "UpgradeUnsupported"
 	ConditionReasonComponentCreationFailed               ConditionReason = "ComponentCreationFailed"
 	ConditionReasonRouteComponentDependencyError         ConditionReason = "RouteComponentDependencyError"
 	ConditionReasonObjectStorageComponentDependencyError ConditionReason = "ObjectStorageComponentDependencyError"
+	ConditionReasonMonitoringComponentDependencyError    ConditionReason = "MonitoringComponentDependencyError"
 	ConditionReasonConfigInvalid                         ConditionReason = "ConfigInvalid"
 )
 
@@ -216,7 +250,13 @@ func ComponentsMatch(firstComponents, secondComponents []Component) bool {
 }
 
 // ComponentIsManaged returns whether the given component is managed or not.
-func ComponentIsManaged(components []Component, name string) bool {
+func ComponentIsManaged(components []Component, name ComponentKind) bool {
+	// we do not expose Base component through the CRD, as it represents the base deployment
+	// of a quay registry and will always be managed.
+	if name == ComponentBase {
+		return true
+	}
+
 	for _, c := range components {
 		if c.Kind == name {
 			return c.Managed
@@ -226,7 +266,7 @@ func ComponentIsManaged(components []Component, name string) bool {
 }
 
 // RequiredComponent returns whether the given component is required for Quay or not.
-func RequiredComponent(component string) bool {
+func RequiredComponent(component ComponentKind) bool {
 	for _, c := range requiredComponents {
 		if c == component {
 			return true
@@ -235,40 +275,55 @@ func RequiredComponent(component string) bool {
 	return false
 }
 
-// EnsureDefaultComponents adds any `Components` which are missing from `Spec.Components`
-// and returns a new `QuayRegistry` copy.
+// EnsureDefaultComponents adds any `Components` which are missing from `Spec.Components`.
+// Returns an error if a component was declared as managed but is not supported in the current k8s cluster.
 func EnsureDefaultComponents(ctx *quaycontext.QuayRegistryContext, quay *QuayRegistry) (*QuayRegistry, error) {
 	updatedQuay := quay.DeepCopy()
 	if updatedQuay.Spec.Components == nil {
 		updatedQuay.Spec.Components = []Component{}
 	}
 
-	if ComponentIsManaged(updatedQuay.Spec.Components, "route") && !ctx.SupportsRoutes {
-		return nil, errors.New("cannot use `route` component when `Route` API not available")
+	type componentCheck struct {
+		check func() bool
+		msg   string
+	}
+	componentChecks := map[ComponentKind]componentCheck{
+		ComponentRoute:         {func() bool { return ctx.SupportsRoutes }, "cannot use `route` component when `Route` API not available"},
+		ComponentTLS:           {func() bool { return ctx.SupportsRoutes }, "cannot use `tls` component when `Route` API not available"},
+		ComponentObjectStorage: {func() bool { return ctx.SupportsObjectStorage }, "cannot use `ObjectStorage` component when `ObjectStorage` API not available"},
+		ComponentMonitoring:    {func() bool { return ctx.SupportsMonitoring }, "cannot use `monitoring` component when `Prometheus` API not available"},
 	}
 
-	if ComponentIsManaged(updatedQuay.Spec.Components, "objectstorage") && !ctx.SupportsObjectStorage {
-		return nil, errors.New("cannot use `objectstorage` component when `ObjectBucketClaims` API not available")
+	componentManaged := map[ComponentKind]componentCheck{
+		ComponentTLS: {
+			check: func() bool { return ctx.TLSCert == nil && ctx.TLSKey == nil },
+		},
 	}
 
 	for _, component := range allComponents {
+		componentCheck, checkExists := componentChecks[component]
+		if (checkExists && !componentCheck.check()) && ComponentIsManaged(quay.Spec.Components, component) {
+			return quay, errors.New(componentCheck.msg)
+		}
+
 		found := false
-		for _, definedComponent := range quay.Spec.Components {
-			if component == definedComponent.Kind {
+		for _, declaredComponent := range quay.Spec.Components {
+			if component == declaredComponent.Kind {
 				found = true
 				break
 			}
 		}
 
-		if !found {
-			if component == "route" && !ctx.SupportsRoutes {
-				continue
-			}
-			if component == "objectstorage" && !ctx.SupportsObjectStorage {
-				continue
-			}
+		managed := !checkExists || componentCheck.check()
+		if _, ok := componentManaged[component]; ok {
+			managed = managed && componentManaged[component].check()
+		}
 
-			updatedQuay.Spec.Components = append(updatedQuay.Spec.Components, Component{Kind: component, Managed: true})
+		if !found {
+			updatedQuay.Spec.Components = append(updatedQuay.Spec.Components, Component{
+				Kind:    component,
+				Managed: managed,
+			})
 		}
 	}
 
@@ -291,7 +346,6 @@ func EnsureRegistryEndpoint(ctx *quaycontext.QuayRegistryContext, quay *QuayRegi
 			ctx.ClusterHostname},
 			".")
 	}
-	// TODO: Retrieve load balancer IP from `Service`
 
 	return updatedQuay, quay.Status.RegistryEndpoint == updatedQuay.Status.RegistryEndpoint
 }
@@ -306,11 +360,30 @@ func EnsureConfigEditorEndpoint(ctx *quaycontext.QuayRegistryContext, quay *Quay
 			ctx.ClusterHostname},
 			".")
 	}
-	// TODO: Retrieve load balancer IP from `Service`
 
 	return updatedQuay, quay.Status.ConfigEditorEndpoint == updatedQuay.Status.ConfigEditorEndpoint
 }
 
+// Owns verifies if a QuayRegistry object owns provided Object.
+func Owns(quay QuayRegistry, obj client.Object) bool {
+	for _, owref := range obj.GetOwnerReferences() {
+		if owref.Kind != "QuayRegistry" {
+			continue
+		}
+		if owref.Name != quay.GetName() {
+			continue
+		}
+		if owref.APIVersion != GroupVersion.String() {
+			continue
+		}
+		if owref.UID != quay.UID {
+			continue
+		}
+		return true
+	}
+	return false
+}
+
 // EnsureOwnerReference adds an `ownerReference` to the given object if it does not already have one.
 func EnsureOwnerReference(quay *QuayRegistry, obj client.Object) (client.Object, error) {
 	objectMeta, err := meta.Accessor(obj)
@@ -340,7 +413,24 @@ func EnsureOwnerReference(quay *QuayRegistry, obj client.Object) (client.Object,
 	return obj, nil
 }
 
-const ManagedKeysName = "quay-registry-managed-secret-keys"
+// RemoveOwnerReference removes the `ownerReference` of `QuayRegistry` on the given object.
+func RemoveOwnerReference(quay *QuayRegistry, obj client.Object) (client.Object, error) {
+	filteredOwnerReferences := []metav1.OwnerReference{}
+
+	objectMeta, err := meta.Accessor(obj)
+	if err != nil {
+		return nil, err
+	}
+
+	for _, ownerRef := range objectMeta.GetOwnerReferences() {
+		if ownerRef.Name != quay.GetName() {
+			filteredOwnerReferences = append(filteredOwnerReferences, ownerRef)
+		}
+	}
+	objectMeta.SetOwnerReferences(filteredOwnerReferences)
+
+	return obj, nil
+}
 
 // ManagedKeysSecretNameFor returns the name of the `Secret` in which generated secret keys are stored.
 func ManagedKeysSecretNameFor(quay *QuayRegistry) string {
@@ -351,6 +441,58 @@ func IsManagedKeysSecretFor(quay *QuayRegistry, secret *corev1.Secret) bool {
 	return strings.Contains(secret.GetName(), quay.GetName()+"-"+ManagedKeysName)
 }
 
+func IsManagedTLSSecretFor(quay *QuayRegistry, secret *corev1.Secret) bool {
+	return strings.Contains(secret.GetName(), quay.GetName()+"-"+QuayConfigTLSSecretName)
+}
+
+// FieldGroupNameFor returns the field group name for a component kind.
+func FieldGroupNameFor(cmp ComponentKind) (string, error) {
+	switch cmp {
+	case ComponentClair:
+		return "SecurityScanner", nil
+	case ComponentPostgres:
+		return "Database", nil
+	case ComponentRedis:
+		return "Redis", nil
+	case ComponentObjectStorage:
+		return "DistributedStorage", nil
+	case ComponentRoute:
+		return "HostSettings", nil
+	case ComponentMirror:
+		return "RepoMirror", nil
+	case ComponentHPA:
+		return "", nil
+	case ComponentMonitoring:
+		return "", nil
+	case ComponentTLS:
+		return "", nil
+	default:
+		return "", fmt.Errorf("unknown component: %q", cmp)
+	}
+}
+
+// FieldGroupNamesForManagedComponents returns an slice of group names for all managed components.
+func FieldGroupNamesForManagedComponents(quay *QuayRegistry) ([]string, error) {
+	var fgns []string
+	for _, cmp := range quay.Spec.Components {
+		if !cmp.Managed {
+			continue
+		}
+
+		fgn, err := FieldGroupNameFor(cmp.Kind)
+		if err != nil {
+			return nil, err
+		}
+
+		if len(fgn) == 0 {
+			continue
+		}
+
+		fgns = append(fgns, fgn)
+	}
+	return fgns, nil
+}
+
 func init() {
 	SchemeBuilder.Register(&QuayRegistry{}, &QuayRegistryList{})
 }
