@@ -1,10 +1,15 @@
 package controllers
 
 import (
+	"bytes"
 	"context"
+	"crypto/tls"
+	"encoding/pem"
+	"fmt"
 	"strings"
 	"time"
 
+	prometheusv1 "github.com/coreos/prometheus-operator/pkg/apis/monitoring/v1"
 	objectbucket "github.com/kube-object-storage/lib-bucket-provisioner/pkg/apis/objectbucket.io/v1alpha1"
 	routev1 "github.com/openshift/api/route/v1"
 	"gopkg.in/yaml.v2"
@@ -32,9 +37,17 @@ const (
 
 	databaseSecretKey = "DATABASE_SECRET_KEY"
 	secretKey         = "SECRET_KEY"
+	dbURI             = "DB_URI"
+	dbRootPw          = "DB_ROOT_PW"
+
+	GrafanaDashboardConfigNamespace = "openshift-config-managed"
 )
 
-func (r *QuayRegistryReconciler) checkManagedKeys(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, rawConfig []byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+// FeatureDetection is a method which should return an updated `QuayRegistryContext` after performing a feature detection task.
+// TODO(alecmerdler): Refactor all "feature detection" functions to use a common function interface...
+type FeatureDetection func(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error)
+
+func (r *QuayRegistryReconciler) checkManagedKeys(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
 	var secrets corev1.SecretList
 	listOptions := &client.ListOptions{
 		Namespace: quay.GetNamespace(),
@@ -51,6 +64,8 @@ func (r *QuayRegistryReconciler) checkManagedKeys(ctx *quaycontext.QuayRegistryC
 		if v1.IsManagedKeysSecretFor(quay, &secret) {
 			ctx.DatabaseSecretKey = string(secret.Data[databaseSecretKey])
 			ctx.SecretKey = string(secret.Data[secretKey])
+			ctx.DbUri = string(secret.Data[dbURI])
+			ctx.DbRootPw = string(secret.Data[dbRootPw])
 			break
 		}
 	}
@@ -58,7 +73,40 @@ func (r *QuayRegistryReconciler) checkManagedKeys(ctx *quaycontext.QuayRegistryC
 	return ctx, quay, nil
 }
 
-func (r *QuayRegistryReconciler) checkRoutesAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, rawConfig []byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+func (r *QuayRegistryReconciler) checkManagedTLS(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+	providedTLSCert := configBundle["ssl.cert"]
+	providedTLSKey := configBundle["ssl.key"]
+
+	if providedTLSCert != nil && providedTLSKey != nil {
+		r.Log.Info("provided TLS cert/key pair in `configBundleSecret` will be used")
+		ctx.TLSCert = providedTLSCert
+		ctx.TLSKey = providedTLSKey
+
+		return ctx, quay, nil
+	} else {
+		r.Log.Info("TLS cert/key pair not provided, will use default cluster wildcard cert")
+	}
+
+	return ctx, quay, nil
+}
+
+func (r *QuayRegistryReconciler) checkRoutesAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+	// NOTE: The `route` component is unique because we allow users to set the `SERVER_HOSTNAME` field instead of controlling the entire fieldgroup.
+	// This value is then passed to the created `Route` using a Kustomize variable.
+	var config map[string]interface{}
+	if err := yaml.Unmarshal(configBundle["config.yaml"], &config); err != nil {
+		return ctx, quay, err
+	}
+
+	fieldGroup, err := hostsettings.NewHostSettingsFieldGroup(config)
+	if err != nil {
+		return ctx, quay, err
+	}
+
+	if fieldGroup.ServerHostname != "" {
+		ctx.ServerHostname = fieldGroup.ServerHostname
+	}
+
 	fakeRoute, err := v1.EnsureOwnerReference(quay, &routev1.Route{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      quay.GetName() + "-test-route",
@@ -82,7 +130,10 @@ func (r *QuayRegistryReconciler) checkRoutesAvailable(ctx *quaycontext.QuayRegis
 
 			if len(fakeRoute.(*routev1.Route).Status.Ingress) > 0 {
 				ctx.SupportsRoutes = true
-				ctx.ClusterHostname = fakeRoute.(*routev1.Route).Status.Ingress[0].RouterCanonicalHostname
+				routerName := fakeRoute.(*routev1.Route).Status.Ingress[0].RouterName
+				routerCanonicalHostname := fakeRoute.(*routev1.Route).Status.Ingress[0].RouterCanonicalHostname
+				ctx.ClusterHostname = strings.TrimPrefix(routerCanonicalHostname, "router-"+routerName+".")
+				r.Log.Info("Detected cluster hostname " + ctx.ClusterHostname)
 
 				return true, nil
 			}
@@ -95,21 +146,7 @@ func (r *QuayRegistryReconciler) checkRoutesAvailable(ctx *quaycontext.QuayRegis
 			return ctx, quay, err
 		}
 
-		// NOTE: The `route` component is unique because we allow users to set the `SERVER_HOSTNAME` field instead of controlling the entire fieldgroup.
-		// This value is then passed to the created `Route` using a Kustomize variable.
-		var config map[string]interface{}
-		if err := yaml.Unmarshal(rawConfig, &config); err != nil {
-			return ctx, quay, err
-		}
-
-		fieldGroup, err := hostsettings.NewHostSettingsFieldGroup(config)
-		if err != nil {
-			return ctx, quay, err
-		}
-
-		if fieldGroup.ServerHostname != "" {
-			ctx.ServerHostname = fieldGroup.ServerHostname
-		} else {
+		if ctx.ServerHostname == "" {
 			ctx.ServerHostname = strings.Join([]string{
 				strings.Join([]string{quay.GetName(), "quay", quay.GetNamespace()}, "-"),
 				ctx.ClusterHostname},
@@ -118,6 +155,16 @@ func (r *QuayRegistryReconciler) checkRoutesAvailable(ctx *quaycontext.QuayRegis
 
 		r.Log.Info("detected router canonical hostname: " + ctx.ClusterHostname)
 
+		// TODO(alecmerdler): Try to fetch the wildcard cert from the `ConfigMap` at `openshift-config-managed/default-ingress-cert`...
+		clusterWildcardCert, err := getCertificatesPEM(fakeRoute.(*routev1.Route).Spec.Host + ":443")
+		if err != nil {
+			return ctx, quay, err
+		}
+
+		ctx.ClusterWildcardCert = clusterWildcardCert
+
+		r.Log.Info("detected cluster wildcard certificate for " + ctx.ClusterHostname)
+
 		if err := r.Client.Delete(context.Background(), fakeRoute); err != nil {
 			return ctx, quay, err
 		}
@@ -130,7 +177,7 @@ func (r *QuayRegistryReconciler) checkRoutesAvailable(ctx *quaycontext.QuayRegis
 	return ctx, quay, nil
 }
 
-func (r *QuayRegistryReconciler) checkObjectBucketClaimsAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, rawConfig []byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+func (r *QuayRegistryReconciler) checkObjectBucketClaimsAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
 	datastoreName := types.NamespacedName{Namespace: quay.GetNamespace(), Name: quay.GetName() + "-quay-datastore"}
 	var objectBucketClaims objectbucket.ObjectBucketClaimList
 	if err := r.Client.List(context.Background(), &objectBucketClaims); err == nil {
@@ -186,9 +233,9 @@ func (r *QuayRegistryReconciler) checkObjectBucketClaimsAvailable(ctx *quayconte
 }
 
 // TODO: Improve this once `builds` is a managed component.
-func (r *QuayRegistryReconciler) checkBuildManagerAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, rawConfig []byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+func (r *QuayRegistryReconciler) checkBuildManagerAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
 	var config map[string]interface{}
-	if err := yaml.Unmarshal(rawConfig, &config); err != nil {
+	if err := yaml.Unmarshal(configBundle["config.yaml"], &config); err != nil {
 		return ctx, quay, err
 	}
 
@@ -199,6 +246,45 @@ func (r *QuayRegistryReconciler) checkBuildManagerAvailable(ctx *quaycontext.Qua
 	return ctx, quay, nil
 }
 
+// Validates if the monitoring component can be run. We assume that we are
+// running in an Openshift environment with cluster monitoring enabled for our
+// monitoring component to work
+func (r *QuayRegistryReconciler) checkMonitoringAvailable(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, configBundle map[string][]byte) (*quaycontext.QuayRegistryContext, *v1.QuayRegistry, error) {
+	if len(r.WatchNamespace) > 0 {
+		msg := "monitoring is only supported in AllNamespaces mode. Disabling component monitoring"
+		r.Log.Info(msg)
+		err := fmt.Errorf(msg)
+		return ctx, quay, err
+	}
+
+	var serviceMonitors prometheusv1.ServiceMonitorList
+	if err := r.Client.List(context.Background(), &serviceMonitors); err != nil {
+		r.Log.Info("Unable to find ServiceMonitor CRD. Monitoring component disabled")
+		return ctx, quay, err
+	}
+	r.Log.Info("cluster supports `ServiceMonitor` API")
+
+	var prometheusRules prometheusv1.PrometheusRuleList
+	if err := r.Client.List(context.Background(), &prometheusRules); err != nil {
+		r.Log.Info("Unable to find PrometheusRule CRD. Monitoring component disabled")
+		return ctx, quay, err
+	}
+	r.Log.Info("cluster supports `PrometheusRules` API")
+
+	namespaceKey := types.NamespacedName{Name: GrafanaDashboardConfigNamespace}
+	var grafanaDashboardNamespace corev1.Namespace
+	if err := r.Client.Get(context.Background(), namespaceKey, &grafanaDashboardNamespace); err != nil {
+		msg := fmt.Sprintf("Unable to find the Grafana config namespace %s. Monitoring component disabled", GrafanaDashboardConfigNamespace)
+		r.Log.Info(msg)
+		return ctx, quay, err
+	}
+	r.Log.Info(GrafanaDashboardConfigNamespace + " found")
+
+	ctx.SupportsMonitoring = true
+
+	return ctx, quay, nil
+}
+
 func configEditorCredentialsSecretFrom(objs []client.Object) string {
 	for _, obj := range objs {
 		objectMeta, _ := meta.Accessor(obj)
@@ -212,3 +298,26 @@ func configEditorCredentialsSecretFrom(objs []client.Object) string {
 
 	return ""
 }
+
+// Taken from https://stackoverflow.com/questions/46735347/how-can-i-fetch-a-certificate-from-a-url
+func getCertificatesPEM(address string) ([]byte, error) {
+	conn, err := tls.Dial("tcp", address, &tls.Config{
+		InsecureSkipVerify: true,
+	})
+	if err != nil {
+		return nil, err
+	}
+	defer conn.Close()
+	var b bytes.Buffer
+	for _, cert := range conn.ConnectionState().PeerCertificates {
+		err := pem.Encode(&b, &pem.Block{
+			Type:  "CERTIFICATE",
+			Bytes: cert.Raw,
+		})
+		if err != nil {
+			return nil, err
+		}
+	}
+
+	return b.Bytes(), nil
+}
