@@ -4,21 +4,24 @@ import (
 	"context"
 	"crypto/rand"
 	"encoding/hex"
+	"sync"
 	"time"
 
 	. "github.com/onsi/ginkgo"
 	. "github.com/onsi/gomega"
 	appsv1 "k8s.io/api/apps/v1"
-	batchv1 "k8s.io/api/batch/v1"
 	corev1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/client-go/kubernetes/scheme"
+	"k8s.io/client-go/util/cert"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/reconcile"
 
 	v1 "github.com/quay/quay-operator/apis/quay/v1"
 	quaycontext "github.com/quay/quay-operator/pkg/context"
+	"github.com/quay/quay-operator/pkg/kustomize"
 )
 
 func newQuayRegistry(name, namespace string) *v1.QuayRegistry {
@@ -35,18 +38,40 @@ func newQuayRegistry(name, namespace string) *v1.QuayRegistry {
 	}
 	// TODO: Test omitting components and marking some as disabled/unmanaged...
 	quay, _ = v1.EnsureDefaultComponents(
-		&quaycontext.QuayRegistryContext{SupportsRoutes: false, SupportsObjectStorage: false},
+		&quaycontext.QuayRegistryContext{SupportsRoutes: false, SupportsObjectStorage: false, SupportsMonitoring: false},
 		quay.DeepCopy(),
 	)
 
 	return quay
 }
 
-func newConfigBundle(name, namespace string) corev1.Secret {
+func newConfigBundle(name, namespace string, withCerts bool) corev1.Secret {
 	config := map[string]interface{}{
 		"ENTERPRISE_LOGO_URL": "/static/img/quay-horizontal-color.svg",
 		"FEATURE_SUPER_USERS": true,
 		"SERVER_HOSTNAME":     "quay-app.quay-enterprise",
+		// Since the testing cluster doesn't support storage, we must mock an unmanaged storage
+		"DISTRIBUTED_STORAGE_CONFIG":            map[string]interface{}{},
+		"DISTRIBUTED_STORAGE_DEFAULT_LOCATIONS": []string{},
+		"DISTRIBUTED_STORAGE_PREFERENCE":        []string{},
+	}
+
+	data := map[string][]byte{
+		"config.yaml": encode(config),
+	}
+
+	if withCerts {
+		cert, key, err := cert.GenerateSelfSignedCertKey(
+			config["SERVER_HOSTNAME"].(string),
+			nil,
+			[]string{config["SERVER_HOSTNAME"].(string)})
+
+		if err != nil {
+			panic(err)
+		}
+
+		data["ssl.cert"] = cert
+		data["ssl.key"] = key
 	}
 
 	return corev1.Secret{
@@ -54,9 +79,7 @@ func newConfigBundle(name, namespace string) corev1.Secret {
 			Name:      name,
 			Namespace: namespace,
 		},
-		Data: map[string][]byte{
-			"config.yaml": encode(config),
-		},
+		Data: data,
 	}
 }
 
@@ -83,27 +106,16 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 		Expect(refs[0].Name).To(Equal(quayRegistry.GetName()))
 	}
 
-	// progressUpgradeDeployment sets the `status` manually because `envtest` only runs apiserver, not controllers.
-	progressUpgradeJob := func() error {
-		var upgradeJob batchv1.Job
-		err := k8sClient.Get(context.Background(), types.NamespacedName{Name: quayRegistry.GetName() + "-quay-app-upgrade", Namespace: namespace}, &upgradeJob)
-		if err != nil {
-			return nil
-		}
-
-		upgradeJob.Status.Succeeded = 1
-
-		return k8sClient.Status().Update(context.Background(), &upgradeJob)
-	}
-
 	BeforeEach(func() {
 		namespace = randIdentifier(16)
 
+		var mtx sync.Mutex
 		controller = &QuayRegistryReconciler{
 			Client:        k8sClient,
 			Log:           testLogger,
 			Scheme:        scheme.Scheme,
 			EventRecorder: testEventRecorder,
+			Mtx:           &mtx,
 		}
 
 		Expect(k8sClient.Create(context.Background(), &corev1.Namespace{ObjectMeta: metav1.ObjectMeta{Name: namespace}})).Should(Succeed())
@@ -139,7 +151,7 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 					Name:      updatedQuayRegistry.Spec.ConfigBundleSecret,
 					Namespace: quayRegistry.GetNamespace()},
 				&configBundleSecret)).
-				Should(Succeed())
+				To(Succeed())
 		})
 
 		It("will reference the same `configBundleSecret` when reconciled again", func() {
@@ -163,8 +175,38 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 					Name:      updatedQuayRegistry.Spec.ConfigBundleSecret,
 					Namespace: quayRegistry.GetNamespace()},
 				&configBundleSecret)).
-				Should(Succeed())
+				To(Succeed())
 		})
+
+		// This test needs to be fixed. Since the ObjectStorage API is not available, we cannot use managed storage.
+		// Since the generated config bundle does not have fields for object storage, this secret is never created.
+		// In order to properly test this behavior, we must transition our testing into a live cluster (see TODOs)
+
+		// It("should not generate a self-signed TLS cert/key pair in a new `Secret`", func() {
+		// 	// Reconcile again to get past defaulting step
+		// 	result, err = controller.Reconcile(context.Background(), reconcile.Request{NamespacedName: quayRegistryName})
+
+		// 	var secrets corev1.SecretList
+		// 	listOptions := client.ListOptions{
+		// 		Namespace: namespace,
+		// 		LabelSelector: labels.SelectorFromSet(map[string]string{
+		// 			kustomize.QuayRegistryNameLabel: quayRegistryName.Name,
+		// 		})}
+
+		// 	Expect(k8sClient.List(context.Background(), &secrets, &listOptions)).To(Succeed())
+
+		// 	found := false
+		// 	for _, secret := range secrets.Items {
+		// 		if v1.IsManagedTLSSecretFor(quayRegistry, &secret) {
+		// 			found = true
+
+		// 			Expect(secret.Data).NotTo(HaveKey("ssl.cert"))
+		// 			Expect(secret.Data).NotTo(HaveKey("ssl.key"))
+		// 		}
+		// 	}
+
+		// 	Expect(found).To(BeTrue())
+		// })
 	})
 
 	When("it references a `configBundleSecret` that does not exist", func() {
@@ -211,7 +253,7 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 	When("it references a `configBundleSecret` that does exist", func() {
 		BeforeEach(func() {
 			quayRegistry = newQuayRegistry("test-registry", namespace)
-			configBundle = newConfigBundle("quay-config-secret-abc123", namespace)
+			configBundle = newConfigBundle("quay-config-secret-abc123", namespace, true)
 			quayRegistry.Spec.ConfigBundleSecret = configBundle.GetName()
 			quayRegistryName = types.NamespacedName{
 				Name:      quayRegistry.Name,
@@ -253,8 +295,6 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 		})
 
 		It("reports the current version in the `status` block", func() {
-			Expect(progressUpgradeJob()).To(Succeed())
-
 			var updatedQuayRegistry v1.QuayRegistry
 
 			Eventually(func() v1.QuayVersion {
@@ -262,12 +302,37 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 				return updatedQuayRegistry.Status.CurrentVersion
 			}, time.Second*30).Should(Equal(v1.QuayVersionCurrent))
 		})
+
+		It("should copy the provided TLS cert/key pair into a new `Secret`", func() {
+			var secrets corev1.SecretList
+			listOptions := client.ListOptions{
+				Namespace: namespace,
+				LabelSelector: labels.SelectorFromSet(map[string]string{
+					kustomize.QuayRegistryNameLabel: quayRegistryName.Name,
+				})}
+
+			Expect(k8sClient.List(context.Background(), &secrets, &listOptions)).To(Succeed())
+
+			found := false
+			for _, secret := range secrets.Items {
+				if v1.IsManagedTLSSecretFor(quayRegistry, &secret) {
+					found = true
+
+					Expect(secret.Data).To(HaveKey("ssl.cert"))
+					Expect(secret.Data["ssl.cert"]).To(Equal(configBundle.Data["ssl.cert"]))
+					Expect(secret.Data).To(HaveKey("ssl.key"))
+					Expect(secret.Data["ssl.key"]).To(Equal(configBundle.Data["ssl.key"]))
+				}
+			}
+
+			Expect(found).To(BeTrue())
+		})
 	})
 
 	When("the current version in the `status` block is the same as the Operator", func() {
 		BeforeEach(func() {
 			quayRegistry = newQuayRegistry("test-registry", namespace)
-			configBundle = newConfigBundle("quay-config-secret-abc123", namespace)
+			configBundle = newConfigBundle("quay-config-secret-abc123", namespace, true)
 			quayRegistry.Spec.ConfigBundleSecret = configBundle.GetName()
 			quayRegistryName = types.NamespacedName{
 				Name:      quayRegistry.Name,
@@ -297,7 +362,7 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 	When("the current version in the `status` block is upgradable", func() {
 		BeforeEach(func() {
 			quayRegistry = newQuayRegistry("test-registry", namespace)
-			configBundle = newConfigBundle("quay-config-secret-abc123", namespace)
+			configBundle = newConfigBundle("quay-config-secret-abc123", namespace, true)
 			quayRegistry.Spec.ConfigBundleSecret = configBundle.GetName()
 			quayRegistryName = types.NamespacedName{
 				Name:      quayRegistry.Name,
@@ -307,15 +372,12 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 			Expect(k8sClient.Create(context.Background(), &configBundle)).Should(Succeed())
 			Expect(k8sClient.Create(context.Background(), quayRegistry)).Should(Succeed())
 
-			quayRegistry.Status.CurrentVersion = v1.QuayVersionPrevious
-
 			Expect(k8sClient.Status().Update(context.Background(), quayRegistry)).To(Succeed())
 
 			result, err = controller.Reconcile(context.Background(), reconcile.Request{NamespacedName: quayRegistryName})
 		})
 
 		It("successfully performs an upgrade", func() {
-			Expect(progressUpgradeJob()).To(Succeed())
 			Expect(err).NotTo(HaveOccurred())
 			Expect(result.Requeue).To(BeFalse())
 
@@ -363,7 +425,7 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 				Name:      quayRegistry.Name,
 				Namespace: quayRegistry.Namespace,
 			}
-			configBundle = newConfigBundle("quay-config-secret-abc123", namespace)
+			configBundle = newConfigBundle("quay-config-secret-abc123", namespace, true)
 			quayRegistry.Spec.ConfigBundleSecret = configBundle.GetName()
 			quayRegistry.Spec.Components = nil
 
@@ -384,7 +446,7 @@ var _ = Describe("Reconciling a QuayRegistry", func() {
 			Expect(k8sClient.Get(context.Background(), quayRegistryName, &updatedQuayRegistry)).Should(Succeed())
 
 			expectedQuay, err := v1.EnsureDefaultComponents(
-				&quaycontext.QuayRegistryContext{SupportsRoutes: false, SupportsObjectStorage: false},
+				&quaycontext.QuayRegistryContext{SupportsRoutes: false, SupportsObjectStorage: false, SupportsMonitoring: false},
 				quayRegistry.DeepCopy(),
 			)
 
