@@ -7,8 +7,10 @@ import (
 	"os"
 	"path/filepath"
 	"runtime"
+	"sort"
 	"strings"
 
+	prometheusv1 "github.com/coreos/prometheus-operator/pkg/apis/monitoring/v1"
 	"github.com/go-logr/logr"
 	objectbucket "github.com/kube-object-storage/lib-bucket-provisioner/pkg/apis/objectbucket.io/v1alpha1"
 	route "github.com/openshift/api/route/v1"
@@ -16,8 +18,7 @@ import (
 	autoscaling "k8s.io/api/autoscaling/v2beta2"
 	batchv1 "k8s.io/api/batch/v1"
 	corev1 "k8s.io/api/core/v1"
-	rbac "k8s.io/api/rbac/v1beta1"
-	"k8s.io/apimachinery/pkg/api/meta"
+	rbac "k8s.io/api/rbac/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/apimachinery/pkg/util/json"
@@ -30,6 +31,7 @@ import (
 
 	v1 "github.com/quay/quay-operator/apis/quay/v1"
 	quaycontext "github.com/quay/quay-operator/pkg/context"
+	"github.com/quay/quay-operator/pkg/middleware"
 )
 
 const (
@@ -38,7 +40,6 @@ const (
 	configSecretPrefix                = "quay-config-secret"
 	registryHostnameAnnotation        = "quay-registry-hostname"
 	buildManagerHostnameAnnotation    = "quay-buildmanager-hostname"
-	managedFieldGroupsAnnotation      = "quay-managed-fieldgroups"
 	operatorServiceEndpointAnnotation = "quay-operator-service-endpoint"
 
 	podNamespaceKey = "MY_POD_NAMESPACE"
@@ -50,18 +51,18 @@ const (
 
 // componentImageFor checks for an environment variable indicating which component container image
 // to use. If set, returns a Kustomize image override for the given component.
-func componentImageFor(component string) types.Image {
-	envVarFor := map[string]string{
-		"base":     componentImagePrefix + "QUAY",
-		"clair":    componentImagePrefix + "CLAIR",
-		"redis":    componentImagePrefix + "REDIS",
-		"postgres": componentImagePrefix + "POSTGRES",
+func componentImageFor(component v1.ComponentKind) (types.Image, error) {
+	envVarFor := map[v1.ComponentKind]string{
+		v1.ComponentBase:     componentImagePrefix + "QUAY",
+		v1.ComponentClair:    componentImagePrefix + "CLAIR",
+		v1.ComponentRedis:    componentImagePrefix + "REDIS",
+		v1.ComponentPostgres: componentImagePrefix + "POSTGRES",
 	}
-	defaultImagesFor := map[string]string{
-		"base":     "quay.io/projectquay/quay",
-		"clair":    "quay.io/projectquay/clair",
-		"redis":    "centos/redis-32-centos7",
-		"postgres": "centos/postgresql-10-centos7",
+	defaultImagesFor := map[v1.ComponentKind]string{
+		v1.ComponentBase:     "quay.io/projectquay/quay",
+		v1.ComponentClair:    "quay.io/projectquay/clair",
+		v1.ComponentRedis:    "centos/redis-32-centos7",
+		v1.ComponentPostgres: "centos/postgresql-10-centos7",
 	}
 
 	imageOverride := types.Image{
@@ -69,15 +70,18 @@ func componentImageFor(component string) types.Image {
 	}
 	image := os.Getenv(envVarFor[component])
 	if image != "" {
-		if len(strings.Split(image, "@")) != 2 {
-			panic(envVarFor[component] + " must use manifest digest reference")
+		if len(strings.Split(image, "@")) == 2 {
+			imageOverride.NewName = strings.Split(image, "@")[0]
+			imageOverride.Digest = strings.Split(image, "@")[1]
+		} else if len(strings.Split(image, ":")) == 2 {
+			imageOverride.NewName = strings.Split(image, ":")[0]
+			imageOverride.NewTag = strings.Split(image, ":")[1]
+		} else {
+			return types.Image{}, errors.New("Image override must be reference by tag or manifest digest: " + image)
 		}
-
-		imageOverride.NewName = strings.Split(image, "@")[0]
-		imageOverride.Digest = strings.Split(image, "@")[1]
 	}
 
-	return imageOverride
+	return imageOverride, nil
 }
 
 func kustomizeDir() string {
@@ -99,6 +103,10 @@ func upgradeOverlayDir() string {
 	return filepath.Join(kustomizeDir(), "overlays", "current", "upgrade")
 }
 
+func unmanagedTLSOverlayDir() string {
+	return filepath.Join(kustomizeDir(), "overlays", "current", "unmanaged-tls")
+}
+
 func configEditorOnlyOverlay() string {
 	return filepath.Join(kustomizeDir(), "overlays", "current", "config-only")
 }
@@ -129,23 +137,37 @@ func decode(bytes []byte) interface{} {
 	return value
 }
 
+// EnsureCreationOrder sorts the given slice of Kubernetes objects so that
+// when created in order, `Deployments` will be created after their dependencies (`Secrets`/`ConfigMaps`/etc).
+func EnsureCreationOrder(objects []client.Object) []client.Object {
+	sort.Slice(objects, func(i, j int) bool {
+		return objects[i].GetObjectKind().GroupVersionKind().String() != schema.GroupVersionKind{Group: "apps", Version: "v1", Kind: "Deployment"}.String()
+	})
+
+	return objects
+}
+
 // ModelFor returns an empty Kubernetes object instance for the given `GroupVersionKind`.
 // Example: Calling with `core.v1.Secret` GVK returns an empty `corev1.Secret` instance.
 func ModelFor(gvk schema.GroupVersionKind) client.Object {
 	switch gvk.String() {
+	case schema.GroupVersionKind{Version: "v1", Kind: "Namespace"}.String():
+		return &corev1.Namespace{}
 	case schema.GroupVersionKind{Version: "v1", Kind: "Secret"}.String():
 		return &corev1.Secret{}
 	case schema.GroupVersionKind{Version: "v1", Kind: "Service"}.String():
 		return &corev1.Service{}
 	case schema.GroupVersionKind{Version: "v1", Kind: "ConfigMap"}.String():
 		return &corev1.ConfigMap{}
+	case schema.GroupVersionKind{Version: "v1", Kind: "ServiceAccount"}.String():
+		return &corev1.ServiceAccount{}
 	case schema.GroupVersionKind{Version: "v1", Kind: "PersistentVolumeClaim"}.String():
 		return &corev1.PersistentVolumeClaim{}
 	case schema.GroupVersionKind{Group: "apps", Version: "v1", Kind: "Deployment"}.String():
 		return &apps.Deployment{}
-	case schema.GroupVersionKind{Group: "rbac.authorization.k8s.io", Version: "v1beta1", Kind: "Role"}.String():
+	case schema.GroupVersionKind{Group: "rbac.authorization.k8s.io", Version: "v1", Kind: "Role"}.String():
 		return &rbac.Role{}
-	case schema.GroupVersionKind{Group: "rbac.authorization.k8s.io", Version: "v1beta1", Kind: "RoleBinding"}.String():
+	case schema.GroupVersionKind{Group: "rbac.authorization.k8s.io", Version: "v1", Kind: "RoleBinding"}.String():
 		return &rbac.RoleBinding{}
 	case schema.GroupVersionKind{Group: "route.openshift.io", Version: "v1", Kind: "Route"}.String():
 		return &route.Route{}
@@ -155,6 +177,10 @@ func ModelFor(gvk schema.GroupVersionKind) client.Object {
 		return &autoscaling.HorizontalPodAutoscaler{}
 	case schema.GroupVersionKind{Group: "batch", Version: "v1", Kind: "Job"}.String():
 		return &batchv1.Job{}
+	case schema.GroupVersionKind{Group: "monitoring.coreos.com", Version: "v1", Kind: "ServiceMonitor"}.String():
+		return &prometheusv1.ServiceMonitor{}
+	case schema.GroupVersionKind{Group: "monitoring.coreos.com", Version: "v1", Kind: "PrometheusRule"}.String():
+		return &prometheusv1.PrometheusRule{}
 	default:
 		panic(fmt.Sprintf("Missing model for GVK %s", gvk.String()))
 	}
@@ -241,6 +267,25 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 		return nil, err
 	}
 
+	if ctx.DbRootPw == "" {
+		rootpw, err := generateRandomString(32)
+		if err != nil {
+			return nil, err
+		}
+		ctx.DbRootPw = rootpw
+	}
+
+	quayConfigTLSSources := []string{}
+	if ctx.ClusterWildcardCert != nil {
+		quayConfigTLSSources = append(quayConfigTLSSources, "ocp-cluster-wildcard.cert="+string(ctx.ClusterWildcardCert))
+	}
+	if ctx.TLSCert != nil {
+		quayConfigTLSSources = append(quayConfigTLSSources, "ssl.cert="+string(ctx.TLSCert))
+	}
+	if ctx.TLSKey != nil {
+		quayConfigTLSSources = append(quayConfigTLSSources, "ssl.key="+string(ctx.TLSKey))
+	}
+
 	generatedSecrets := []types.SecretArgs{
 		{
 			GeneratorArgs: types.GeneratorArgs{
@@ -257,10 +302,20 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 					LiteralSources: []string{
 						"DATABASE_SECRET_KEY=" + ctx.DatabaseSecretKey,
 						"SECRET_KEY=" + ctx.SecretKey,
+						"DB_URI=" + ctx.DbUri,
+						"DB_ROOT_PW=" + ctx.DbRootPw,
 					},
 				},
 			},
 		},
+		{
+			GeneratorArgs: types.GeneratorArgs{
+				Name: v1.QuayConfigTLSSecretName,
+				KvPairSources: types.KvPairSources{
+					LiteralSources: quayConfigTLSSources,
+				},
+			},
+		},
 		{
 			GeneratorArgs: types.GeneratorArgs{
 				Name: "quay-config-editor-credentials",
@@ -275,13 +330,11 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 	}
 
 	componentPaths := []string{}
-	managedFieldGroups := []string{}
 	for _, component := range quay.Spec.Components {
 		if component.Managed {
-			componentPaths = append(componentPaths, filepath.Join("..", "components", component.Kind))
-			managedFieldGroups = append(managedFieldGroups, fieldGroupNameFor(component.Kind))
+			componentPaths = append(componentPaths, filepath.Join("..", "components", string(component.Kind)))
 
-			componentConfigFiles, err := componentConfigFilesFor(component.Kind, quay, quayConfigFiles)
+			componentConfigFiles, err := componentConfigFilesFor(ctx, component.Kind, quay, quayConfigFiles)
 			if componentConfigFiles == nil || err != nil {
 				continue
 			}
@@ -293,7 +346,7 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 
 			generatedSecrets = append(generatedSecrets, types.SecretArgs{
 				GeneratorArgs: types.GeneratorArgs{
-					Name:     component.Kind + "-config-secret",
+					Name:     string(component.Kind) + "-config-secret",
 					Behavior: "merge",
 					KvPairSources: types.KvPairSources{
 						LiteralSources: sources,
@@ -305,11 +358,12 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 
 	images := []types.Image{}
 	for _, component := range append(quay.Spec.Components, v1.Component{Kind: "base", Managed: true}) {
-		if component.Managed {
-			imageOverride := componentImageFor(component.Kind)
-			if imageOverride.Digest != "" {
-				images = append(images, imageOverride)
-			}
+		image, err := componentImageFor(component.Kind)
+		if err != nil {
+			return nil, err
+		}
+		if image.NewName != "" || image.Digest != "" {
+			images = append(images, image)
 		}
 	}
 
@@ -328,7 +382,6 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 			QuayRegistryNameLabel: quay.GetName(),
 		},
 		CommonAnnotations: map[string]string{
-			managedFieldGroupsAnnotation:      strings.ReplaceAll(strings.Join(managedFieldGroups, ","), ",,", ","),
 			registryHostnameAnnotation:        ctx.ServerHostname,
 			buildManagerHostnameAnnotation:    strings.Split(ctx.BuildManagerHostname, ":")[0],
 			operatorServiceEndpointAnnotation: operatorServiceEndpoint(),
@@ -349,43 +402,6 @@ func KustomizationFor(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistr
 	}, nil
 }
 
-// flattenSecret takes all Quay config fields in given secret and combines them under `config.yaml` key.
-func flattenSecret(configBundle *corev1.Secret) (*corev1.Secret, error) {
-	flattenedSecret := configBundle.DeepCopy()
-
-	var flattenedConfig map[string]interface{}
-	if err := yaml.Unmarshal(configBundle.Data["config.yaml"], &flattenedConfig); err != nil {
-		return nil, err
-	}
-
-	isConfigField := func(field string) bool {
-		return strings.Contains(field, ".config.yaml")
-	}
-
-	for key, file := range configBundle.Data {
-		if isConfigField(key) {
-			var valueYAML map[string]interface{}
-			if err := yaml.Unmarshal(file, &valueYAML); err != nil {
-				return nil, err
-			}
-
-			for configKey, configValue := range valueYAML {
-				flattenedConfig[configKey] = configValue
-			}
-			delete(flattenedSecret.Data, key)
-		}
-	}
-
-	flattenedConfigYAML, err := yaml.Marshal(flattenedConfig)
-	if err != nil {
-		return nil, err
-	}
-
-	flattenedSecret.Data["config.yaml"] = []byte(flattenedConfigYAML)
-
-	return flattenedSecret, nil
-}
-
 // Inflate takes a `QuayRegistry` object and returns a set of Kubernetes objects representing a Quay deployment.
 func Inflate(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, baseConfigBundle *corev1.Secret, log logr.Logger) ([]client.Object, error) {
 	// Each managed component brings its own generated `config.yaml` fields
@@ -399,44 +415,51 @@ func Inflate(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, baseCo
 
 	// Generate or pull out the `SECRET_KEY` and `DATABASE_SECRET_KEY`. Since these must be stable across
 	// runs of the same config, we store them (and re-read them) from a specialized `Secret`.
-	if ctx.DatabaseSecretKey == "" {
-		if key, found := parsedUserConfig["DATABASE_SECRET_KEY"]; found {
-			log.Info("`DATABASE_SECRET_KEY` found in user-provided config")
-
-			ctx.DatabaseSecretKey = key.(string)
-		} else {
-			log.Info("`DATABASE_SECRET_KEY` not found in user-provided config, generating new one")
-
-			databaseSecretKey, err := generateRandomString(secretKeyLength)
-			if err != nil {
-				return nil, err
-			}
-
-			ctx.DatabaseSecretKey = databaseSecretKey
+	// TODO(alecmerdler): Refector these three blocks...
+	if key, found := parsedUserConfig["DATABASE_SECRET_KEY"].(string); found && len(key) > 0 {
+		log.Info("`DATABASE_SECRET_KEY` found in user-provided config")
+		ctx.DatabaseSecretKey = key
+	} else if ctx.DatabaseSecretKey == "" {
+		log.Info("`DATABASE_SECRET_KEY` not found in user-provided config, generating new one")
+		databaseSecretKey, err := generateRandomString(secretKeyLength)
+		if err != nil {
+			return nil, err
 		}
+		ctx.DatabaseSecretKey = databaseSecretKey
 	}
-
-	if ctx.SecretKey == "" {
-		if key, found := parsedUserConfig["SECRET_KEY"]; found {
-			log.Info("`SECRET_KEY` found in user-provided config")
-
-			ctx.SecretKey = key.(string)
-		} else {
-			log.Info("`SECRET_KEY` not found in user-provided config, generating new one")
-
-			secretKey, err := generateRandomString(secretKeyLength)
-			if err != nil {
-				return nil, err
-			}
-
-			ctx.SecretKey = secretKey
-		}
-	}
-
-	parsedUserConfig["SETUP_COMPLETE"] = true
 	parsedUserConfig["DATABASE_SECRET_KEY"] = ctx.DatabaseSecretKey
+
+	if key, found := parsedUserConfig["SECRET_KEY"].(string); found && len(key) > 0 {
+		log.Info("`SECRET_KEY` found in user-provided config")
+		ctx.SecretKey = key
+	} else if ctx.SecretKey == "" {
+		log.Info("`SECRET_KEY` not found in user-provided config, generating new one")
+		secretKey, err := generateRandomString(secretKeyLength)
+		if err != nil {
+			return nil, err
+		}
+		ctx.SecretKey = secretKey
+	}
 	parsedUserConfig["SECRET_KEY"] = ctx.SecretKey
 
+	if dbURI, found := parsedUserConfig["DB_URI"].(string); found && len(dbURI) > 0 {
+		ctx.DbUri = dbURI
+	} else if v1.ComponentIsManaged(quay.Spec.Components, v1.ComponentPostgres) && len(ctx.DbUri) == 0 {
+		log.Info("managed `DB_URI` not found in config, generating new one")
+		user := quay.GetName() + "-quay-database"
+		name := quay.GetName() + "-quay-database"
+		host := quay.GetName() + "-quay-database"
+		port := "5432"
+		password, err := generateRandomString(secretKeyLength)
+		if err != nil {
+			return nil, err
+		}
+		ctx.DbUri = fmt.Sprintf(
+			"postgresql://%s:%s@%s:%s/%s", user, password, host, port, name,
+		)
+	}
+	parsedUserConfig["DB_URI"] = ctx.DbUri
+
 	for field, value := range BaseConfig() {
 		if _, ok := parsedUserConfig[field]; !ok {
 			parsedUserConfig[field] = value
@@ -451,48 +474,55 @@ func Inflate(ctx *quaycontext.QuayRegistryContext, quay *v1.QuayRegistry, baseCo
 				return nil, err
 			}
 
-			componentConfigFiles[component.Kind+".config.yaml"] = encode(fieldGroup)
+			componentConfigFiles[string(component.Kind)+".config.yaml"] = encode(fieldGroup)
 		}
 	}
 
-	log.Info("Ensuring `ssl.cert` and `ssl.key` pair for Quay app TLS")
-	tlsCert, tlsKey, err := EnsureTLSFor(ctx, quay, baseConfigBundle.Data["ssl.cert"], baseConfigBundle.Data["ssl.key"])
+	log.Info("Ensuring TLS cert/key pair for Quay app")
+	tlsCert, tlsKey, err := EnsureTLSFor(ctx, quay)
 	if err != nil {
 		return nil, err
 	}
 
-	componentConfigFiles["ssl.cert"] = tlsCert
-	componentConfigFiles["ssl.key"] = tlsKey
+	ctx.TLSCert = tlsCert
+	ctx.TLSKey = tlsKey
 
 	kustomization, err := KustomizationFor(ctx, quay, componentConfigFiles)
-	check(err)
+	if err != nil {
+		return nil, err
+	}
 
 	var overlay string
 	if rolloutBlocked(quay) {
 		overlay = configEditorOnlyOverlay()
 	} else if quay.Status.CurrentVersion != v1.QuayVersionCurrent {
 		overlay = upgradeOverlayDir()
+	} else if !v1.ComponentIsManaged(quay.Spec.Components, v1.ComponentTLS) {
+		overlay = unmanagedTLSOverlayDir()
 	} else {
 		overlay = overlayDir()
 	}
 	resources, err := generate(kustomization, overlay, componentConfigFiles)
-	check(err)
-
-	for index, resource := range resources {
-		objectMeta, err := meta.Accessor(resource)
-		check(err)
-
-		if strings.Contains(objectMeta.GetName(), configSecretPrefix+"-") {
-			configBundleSecret, err := flattenSecret(resource.(*corev1.Secret))
-			check(err)
-
-			resources[index] = configBundleSecret
-		}
+	if err != nil {
+		return nil, err
 	}
 
-	for _, resource := range resources {
+	for index, resource := range resources {
+		obj, err := middleware.Process(ctx, quay, resource)
+		if err != nil {
+			return nil, err
+		}
+
+		resources[index] = obj
+	}
+
+	for index, resource := range resources {
 		resource, err = v1.EnsureOwnerReference(quay, resource)
-		check(err)
+		if err != nil {
+			return nil, err
+		}
+
+		resources[index] = resource
 	}
 
 	return resources, err
