@@ -2,7 +2,6 @@ package kustomize
 
 import (
 	"fmt"
-	"net/url"
 	"testing"
 
 	"github.com/stretchr/testify/assert"
@@ -41,7 +40,7 @@ func quayRegistry(name string) *v1.QuayRegistry {
 
 var fieldGroupForTests = []struct {
 	name      string
-	component string
+	component v1.ComponentKind
 	quay      *v1.QuayRegistry
 	ctx       quaycontext.QuayRegistryContext
 	expected  shared.FieldGroup
@@ -80,7 +79,9 @@ var fieldGroupForTests = []struct {
 		"postgres",
 		"postgres",
 		quayRegistry("test"),
-		quaycontext.QuayRegistryContext{},
+		quaycontext.QuayRegistryContext{
+			DbUri: "postgresql://test-quay-database:postgres@test-quay-database:5432/test-quay-database",
+		},
 		&database.DatabaseFieldGroup{
 			DbUri: "postgresql://test-quay-database:postgres@test-quay-database:5432/test-quay-database",
 			DbConnectionArgs: &database.DbConnectionArgsStruct{
@@ -131,7 +132,7 @@ var fieldGroupForTests = []struct {
 		},
 		&hostsettings.HostSettingsFieldGroup{
 			ServerHostname:         "test-quay-ns.apps.example.com",
-			ExternalTlsTermination: false,
+			ExternalTlsTermination: true,
 			PreferredUrlScheme:     "https",
 		},
 	},
@@ -168,15 +169,6 @@ func TestFieldGroupFor(t *testing.T) {
 
 			assert.True(len(secscanFieldGroup.SecurityScannerV4PSK) > 0, test.name)
 			secscanFieldGroup.SecurityScannerV4PSK = "abc123"
-		} else if test.name == "postgres" {
-			databaseFieldGroup := fieldGroup.(*database.DatabaseFieldGroup)
-			dbURI, err := url.Parse(databaseFieldGroup.DbUri)
-			assert.Nil(err, test.name)
-
-			password, _ := dbURI.User.Password()
-			assert.True(len(password) > 0, test.name)
-			dbURI.User = url.UserPassword(dbURI.User.Username(), "postgres")
-			databaseFieldGroup.DbUri = dbURI.String()
 		}
 
 		expected, err := yaml.Marshal(test.expected)
@@ -190,7 +182,8 @@ func TestFieldGroupFor(t *testing.T) {
 
 var containsComponentConfigTests = []struct {
 	name          string
-	component     string
+	component     v1.ComponentKind
+	managed       bool
 	rawConfig     string
 	expected      bool
 	expectedError error
@@ -198,6 +191,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"ClairContains",
 		"clair",
+		true,
 		`FEATURE_SECURITY_SCANNER: true`,
 		true,
 		nil,
@@ -205,6 +199,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"ClairDoesNotContain",
 		"clair",
+		true,
 		``,
 		false,
 		nil,
@@ -212,6 +207,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"PostgresContains",
 		"postgres",
+		true,
 		`DB_URI: postgresql://test-quay-database:postgres@test-quay-database:5432/test-quay-database`,
 		true,
 		nil,
@@ -219,6 +215,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"PostgresDoesNotContain",
 		"postgres",
+		true,
 		``,
 		false,
 		nil,
@@ -226,6 +223,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"RedisContains",
 		"redis",
+		true,
 		`BUILDLOGS_REDIS:
   host: test-quay-redis
 `,
@@ -235,6 +233,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"RedisDoesNotContain",
 		"redis",
+		true,
 		``,
 		false,
 		nil,
@@ -242,6 +241,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"ObjectStorageContains",
 		"objectstorage",
+		true,
 		`DISTRIBUTED_STORAGE_PREFERENCE: 
   - local_us
 `,
@@ -251,6 +251,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"ObjectStorageDoesNotContain",
 		"objectstorage",
+		true,
 		``,
 		false,
 		nil,
@@ -258,6 +259,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"MirrorContains",
 		"mirror",
+		true,
 		`FEATURE_REPO_MIRROR: true`,
 		true,
 		nil,
@@ -265,6 +267,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"MirrorDeosNotContain",
 		"mirror",
+		true,
 		``,
 		false,
 		nil,
@@ -272,6 +275,7 @@ var containsComponentConfigTests = []struct {
 	{
 		"RouteContains",
 		"route",
+		true,
 		`PREFERRED_URL_SCHEME: http`,
 		true,
 		nil,
@@ -279,20 +283,39 @@ var containsComponentConfigTests = []struct {
 	{
 		"RouteContainsServerHostname",
 		"route",
+		true,
 		`SERVER_HOSTNAME: registry.skynet.com`,
-		false,
+		true,
 		nil,
 	},
 	{
 		"RouteDoesNotContain",
 		"route",
+		true,
 		``,
 		false,
 		nil,
 	},
+	{
+		"RouteUnmanagedDoesNotContain",
+		"route",
+		false,
+		``,
+		false,
+		nil,
+	},
+	{
+		"RouteUnmanagedContainsServerHostname",
+		"route",
+		false,
+		`SERVER_HOSTNAME: registry.skynet.com`,
+		true,
+		nil,
+	},
 	{
 		"HorizontalPodAutoscalerDoesNotContain",
 		"horizontalpodautoscaler",
+		true,
 		``,
 		false,
 		nil,
@@ -304,10 +327,11 @@ func TestContainsComponentConfig(t *testing.T) {
 
 	for _, test := range containsComponentConfigTests {
 		var fullConfig map[string]interface{}
+		var certs map[string][]byte
 		err := yaml.Unmarshal([]byte(test.rawConfig), &fullConfig)
 		assert.Nil(err, test.name)
 
-		contains, err := ContainsComponentConfig(fullConfig, test.component)
+		contains, err := ContainsComponentConfig(fullConfig, certs, v1.Component{Kind: test.component, Managed: test.managed})
 
 		if test.expectedError != nil {
 			assert.NotNil(err, test.name)
@@ -398,12 +422,15 @@ func TestEnsureTLSFor(t *testing.T) {
 
 	for _, test := range ensureTLSForTests {
 		quayRegistry := quayRegistry("test")
+
 		quayContext := quaycontext.QuayRegistryContext{
 			ServerHostname:       test.serverHostname,
 			BuildManagerHostname: test.buildManagerHostname,
+			TLSCert:              test.providedCertKeyPair[0],
+			TLSKey:               test.providedCertKeyPair[1],
 		}
 
-		tlsCert, tlsKey, err := EnsureTLSFor(&quayContext, quayRegistry, test.providedCertKeyPair[0], test.providedCertKeyPair[1])
+		tlsCert, tlsKey, err := EnsureTLSFor(&quayContext, quayRegistry)
 
 		assert.Equal(test.expectedErr, err, test.name)
 
@@ -413,10 +440,13 @@ func TestEnsureTLSFor(t *testing.T) {
 				assert.Equal(string(test.providedCertKeyPair[1]), string(tlsKey), test.name)
 			}
 
-			shared.ValidateCertPairWithHostname(tlsCert, tlsKey, test.serverHostname, fieldGroupNameFor("route"))
+			fgn, err := v1.FieldGroupNameFor(v1.ComponentRoute)
+			assert.Nil(err, "error not nil: %s", err)
+
+			shared.ValidateCertPairWithHostname(tlsCert, tlsKey, test.serverHostname, fgn)
 
 			if test.buildManagerHostname != "" {
-				shared.ValidateCertPairWithHostname(tlsCert, tlsKey, test.buildManagerHostname, fieldGroupNameFor("route"))
+				shared.ValidateCertPairWithHostname(tlsCert, tlsKey, test.buildManagerHostname, fgn)
 			}
 		}
 	}
